/*  File UTMATRIX.C  Dynamic matrix allocation utilities.  */
/*  The memory allocation techniques used here are based on */
/*  the public-domain routines in Appendix B of  */
/*  "Numerical Recipes in C", 2nd Ed., by W. Press et al. */
/*  Cambridge University Press (1992) */

#ifndef UTMATRIX_C
#define UTMATRIX_C
#define ARR_END 1
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include "ut.h"

void utmc_error_msg (char *the_msg) {
   textmode (LASTMODE);
   gotoxy (1,5);
   cprintf ("UTMATRIX.C Run time error: \r\n%s\r\n",the_msg);
   getch();
   exit (1);
   }

void utmc_msg (char *the_msg) {
   cprintf ("\n\n\r%s\n",the_msg);
   getch();
   cputs ("\n\n");
   }

float **matrix2d (int rows_lower_lim,int rows_upper_lim,
   int cols_lower_lim,int cols_upper_lim) {
   /* Allocate a 2D float matrix with subscript range */
   /* rows_lower_lim..rows_upper_lim,cols_lower_lim..cols_upper_lim */
   int i,no_of_rows = rows_upper_lim - rows_lower_lim + 1,
       no_of_cols = cols_upper_lim - cols_lower_lim + 1;
   float **a;
   /* Allocate pointers to rows */
   a = (float **) malloc((size_t)((no_of_rows+ARR_END)*sizeof(float*)));
   if (!a) utmc_error_msg ("Memory allocation 1 failed in matrix2d");
   a += ARR_END;
   a -= rows_lower_lim;
   /* Allocate rows and set pointers to them */
   a[rows_lower_lim] = (float *) malloc
     ((size_t)((no_of_rows*no_of_cols + ARR_END)*sizeof(float)));
   if (!a[rows_lower_lim]) utmc_error_msg
      ("Memory allocation 2 failed in matrix2d");
   a[rows_lower_lim] += ARR_END;
   a[rows_lower_lim] -= cols_lower_lim;
   for (i=rows_lower_lim + 1;i <= rows_upper_lim; i++)
       a[i] = a[i-1] + no_of_cols;
   /* Return pointer to array of pointers to rows */
   return a;
   }

float ***matrix3d (int rows_lower_lim,int rows_upper_lim,
   int cols_lower_lim,int cols_upper_lim,int n_lower_lim,int n_upper_lim) {
   /* Allocate a 3D float matrix with subscript range */
   /* rows_lower_lim..rows_upper_lim,cols_lower_lim..cols_upper_lim */
   int i,j,no_of_rows = rows_upper_lim - rows_lower_lim + 1,
       no_of_cols = cols_upper_lim - cols_lower_lim + 1,
       n = n_upper_lim - n_lower_lim + 1;
   float ***a;
   /* Allocate pointers to pointers to rows */
   a = (float ***) malloc((size_t)((no_of_rows+ARR_END)*sizeof(float**)));
   if (!a) utmc_error_msg ("Memory allocation 1 failed in matrix3d");
   a += ARR_END;
   a -= rows_lower_lim;
   /* Allocate pointers to rows and set pointers to them */
   a[rows_lower_lim] = (float **) malloc
     ((size_t)((no_of_rows*no_of_cols + ARR_END)*sizeof(float*)));
   if (!a[rows_lower_lim]) utmc_error_msg
      ("Memory allocation 2 failed in matrix3d");
   a[rows_lower_lim] += ARR_END;
   a[rows_lower_lim] -= cols_lower_lim;
   /* Allocate rows and set pointers to them */
   a[rows_lower_lim][cols_lower_lim] = (float *) malloc
     ((size_t)((no_of_rows*no_of_cols*n + ARR_END)*sizeof(float)));
   if (!a[rows_lower_lim][cols_lower_lim]) utmc_error_msg
      ("Memory allocation 3 failed in matrix3d");
   a[rows_lower_lim][cols_lower_lim] += ARR_END;
   a[rows_lower_lim][cols_lower_lim] -= cols_lower_lim;
   for (j = cols_lower_lim + 1;j <= cols_upper_lim; j++)
       a[rows_lower_lim][j] = a[rows_lower_lim][j-1] + n;
   for (i=rows_lower_lim + 1;i <= rows_upper_lim; i++) {
       a[i] = a[i-1] + no_of_cols;
       a[i][cols_lower_lim] = a[i-1][cols_lower_lim] + no_of_cols*n;
       for (j = cols_lower_lim + 1;j <= cols_upper_lim;j++)
	   a[i][j]=a[i][j-1] + n;
       }
   /* Return pointer to array of pointers to rows */
   return a;
   }

float ****matrix4d (int rows_lower_lim,int rows_upper_lim,
   int cols_lower_lim,int cols_upper_lim,
   int m_lower_lim,int m_upper_lim,
   int n_lower_lim,int n_upper_lim) {
   /* Allocate a 4D float matrix with subscript range */
   /* rows_lower_lim..rows_upper_lim,cols_lower_lim..cols_upper_lim */
   /* m_lower_lim..m_upper_lim,n_lower_lim..n_upper_lim  */
   int i,j,k,no_of_rows = rows_upper_lim - rows_lower_lim + 1,
       no_of_cols = cols_upper_lim - cols_lower_lim + 1,
       m = m_upper_lim - m_lower_lim + 1,
       n = n_upper_lim - n_lower_lim + 1;
   float ****a;
   /*  Think of 4d matrix as a 3d array of pointers pointing to */
   /*  vectors of length n */
   /*  Allocate pointers to vectors of float** of length #cols   */
   a = (float ****) malloc((size_t)((no_of_rows+ARR_END)*sizeof(float***)));
   if (!a) utmc_error_msg ("Memory allocation 1 failed in matrix4d");
   a += ARR_END;
   a -= rows_lower_lim;
   /*  Allocate pointers to vectors of float* of length m */
   a[rows_lower_lim] = (float ***) malloc
     ((size_t)((no_of_rows*no_of_cols + ARR_END)*sizeof(float**)));
   if (!a[rows_lower_lim]) utmc_error_msg
      ("Memory allocation 2 failed in matrix4d");
   a[rows_lower_lim] += ARR_END;
   a[rows_lower_lim] -= cols_lower_lim;
   for (i=rows_lower_lim + 1;i<=rows_upper_lim;i++)
       a[i] = a[i-1] + no_of_cols;
   /* Allocate pointers to vectors of float of length n */
   a[rows_lower_lim][cols_lower_lim] = (float **) malloc
     ((size_t)((no_of_rows*no_of_cols*m + ARR_END)*sizeof(float*)));
   if (!a[rows_lower_lim][cols_lower_lim]) utmc_error_msg
      ("Memory allocation 3 failed in matrix4d");
   a[rows_lower_lim][cols_lower_lim] += ARR_END;
   a[rows_lower_lim][cols_lower_lim] -= m_lower_lim;
   /* First row of 2d matrix: */
   for (j=cols_lower_lim + 1;j<=cols_upper_lim;j++)
       a[rows_lower_lim][j] = a[rows_lower_lim][j-1] + m;
   /* Rest of 2d matrix: */
   for (i=rows_lower_lim + 1;i<=rows_upper_lim;i++)
       for (j=cols_lower_lim;j<=cols_upper_lim;j++)
	   a[i][j] = a[i-1][j] + no_of_cols*m;

   /* Allocate float array */
   a[rows_lower_lim][cols_lower_lim][m_lower_lim] = (float *) malloc
     ((size_t)((no_of_rows*no_of_cols*m*n + ARR_END)*sizeof(float)));
   if (!a[rows_lower_lim][cols_lower_lim][m_lower_lim]) utmc_error_msg
      ("Memory allocation 4 failed in matrix4d");
   a[rows_lower_lim][cols_lower_lim][m_lower_lim] += ARR_END;
   a[rows_lower_lim][cols_lower_lim][m_lower_lim] -= n_lower_lim;
   /* Set addresses for 3d array */
   /* Think of 3d array as consisting of 2d layers */
   /* Begin setting addresses in first layer  */
   for (k = m_lower_lim + 1;k <= m_upper_lim; k++)
       a[rows_lower_lim][cols_lower_lim][k] =
	     a[rows_lower_lim][cols_lower_lim][k-1] + n;
   /*  Now set addresses for the rest of the first layer: */
   for (j = cols_lower_lim + 1;j <= cols_upper_lim; j++) {
       a[rows_lower_lim][j][m_lower_lim] =
	     a[rows_lower_lim][j-1][m_lower_lim] + m*n;
       for (k = m_lower_lim + 1;k <= m_upper_lim; k++)
	   a[rows_lower_lim][j][k] = a[rows_lower_lim][j][k-1] + n;
	   }  /* j */
   /* Now do the remaining 2d layers: */
   for (i=rows_lower_lim + 1;i <= rows_upper_lim; i++) {
       a[i][cols_lower_lim][m_lower_lim] =
	  a[i-1][cols_lower_lim][m_lower_lim] + no_of_cols*m*n;
       /* First row of ith layer:  */
       for (k = m_lower_lim + 1;k <= m_upper_lim; k++)
	  a[i][cols_lower_lim][k] = a[i][cols_lower_lim][k-1] + n;
       /* Rest of the ith layer: */
       for (j = cols_lower_lim + 1;j <= cols_upper_lim; j++) {
	  a[i][j][m_lower_lim] = a[i][j-1][m_lower_lim] + m*n;
	  for (k = m_lower_lim + 1;k <= m_upper_lim; k++)
	      a[i][j][k] = a[i][j][k-1] + n;
	      }  /* j */
       }  /* i */
   /* Return pointer to array of pointers to rows */
   return a;
   }

void rand_init_3d_matrix (float ***w,int l_rows,int l_cols,int n_inp,
     float range)
   {
   int i,j,k;
   float r;
   srand(1);
   for (k = 1;k <= n_inp;k++) {
       for (i = 1;i <= l_rows;i++) {
	   for (j = 1;j <= l_cols;j++) {
		r = 2*frand() - 1.0;  /*Random number between -1 and 1*/
		  w [i] [j] [k] = range * r;
		  } /*j*/
	      } /*i*/
	  } /*k*/
   return;
   }

void init_3d_matrix (float ***w,int l_rows,int l_cols,int n_inp,
     const float val)
   {
   int i,j,k;
   float r;
   srand(1);
   for (k = 1;k <= n_inp;k++)
       for (i = 1;i <= l_rows;i++)
	   for (j = 1;j <= l_cols;j++)
	       w [i] [j] [k] = val;
   return;
   }

void rand_init_4d_matrix (float ****w,int l_rows,int l_cols,int m_inp,
   int n_inp, float range)
   /*  Assumes starting index value is 1 for each index */
   {
   int i,j,k,l;
   float r;
   srand(1);
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++)
	    for (k = 1;k <= m_inp;k++)
		 for (l = 1;l <= n_inp;l++) {
		 r = 2*frand() - 1.0;  /*Random number between -1 and 1*/
		 w [i] [j] [k] [l] = range * r;
		 }
   return;
   }

void rand_init_2d_matrix (float **w, int l_rows, int l_cols,float range)
   {
   int i,j;
   float r;
   srand(1);
       for (i = 1;i <= l_rows;i++)
	   for (j = 1;j <= l_cols;j++) {
		r = 2*frand() - 1.0;  /*Random number between -1 and 1*/
		  w [i] [j] = range * r;
		  } /* i,j */
   return;
   }

void init_2d_matrix (float **a, int l_rows, int l_cols, const float val)
   {
   int i,j;
       for (i = 1;i <= l_rows;i++)
	   for (j = 1;j <= l_cols;j++)
		a [i] [j] = val;
   return;
   }

void free_matrix2d (float **a,int rows_lower_lim,int cols_lower_lim) {
     if ((a[rows_lower_lim] + cols_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim] + cols_lower_lim - ARR_END));
     if ((a + rows_lower_lim - ARR_END))
	free ((char *)(a + rows_lower_lim - ARR_END));
     return;
     }

void free_matrix3d (float ***a,int rows_lower_lim,int cols_lower_lim,
     int n_lower_lim) {
     if ((a[rows_lower_lim][cols_lower_lim] + n_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim][cols_lower_lim]
		 + n_lower_lim - ARR_END));
     if ((a[rows_lower_lim] + cols_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim] + cols_lower_lim - ARR_END));
     if ((a + rows_lower_lim - ARR_END))
	free ((char *)(a + rows_lower_lim - ARR_END));
     return;
     }

void free_matrix4d (float ****a,int rows_lower_lim,int cols_lower_lim,
     int m_lower_lim,int n_lower_lim) {
     if ((a[rows_lower_lim][cols_lower_lim][m_lower_lim]
	 + n_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim][cols_lower_lim][m_lower_lim]
	 + n_lower_lim - ARR_END));
     if ((a[rows_lower_lim][cols_lower_lim] + m_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim][cols_lower_lim]
	 + m_lower_lim - ARR_END));
     if ((a[rows_lower_lim] + cols_lower_lim - ARR_END))
	free ((char *)(a[rows_lower_lim] + cols_lower_lim - ARR_END));
     if ((a + rows_lower_lim - ARR_END))
	free ((char *)(a + rows_lower_lim - ARR_END));
     return;
     }

short write_2d_matrix (float **w,int l_rows,int l_cols,char *file_name)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int i,j;
   the_file = fopen (file_name,"wt"); /* Stream file, write only, text */
   if (the_file == 0) return 0;
   fprintf (the_file,"%d  %d\n",l_rows,l_cols);
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++)
	    fprintf (the_file,"%f\n",w[i][j]);
   fclose (the_file);
   return 1;
   }

void print_2d_matrix (float **w,int l_rows,int l_cols)
   {
   int i,j;
   fprintf (stdprn,"%d  %d\n",l_rows,l_cols);
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++)
	    fprintf (stdprn,"%f\n",w[i][j]);
   return;
   }

#define MAX_LINE_LEN 255

short read_2d_matrix (float **w,char *file_name,int rows,int cols)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int l_rows,l_cols;
   int i,j;
   char temp[MAX_LINE_LEN];
   the_file = fopen (file_name,"rt"); /* Stream file, read only, text */
   if (the_file == NULL) {
      utmc_msg("Problem opening 2d matrix file.");
      return 0;
      }
   fgets (temp,MAX_LINE_LEN,the_file);
   if (sscanf (temp,"%d  %d",&l_rows,&l_cols) != 2) {
      utmc_msg ("Problem reading dimensions of 2d matrix.");
      return 0;
      }
   if ((l_rows != rows)||(l_cols != cols)) {
      utmc_msg ("2d matrix file rows and cols do not equal setup values.");
      return 0;
      }
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++) {
	   fgets (temp,MAX_LINE_LEN,the_file);
	   if (sscanf (temp,"%f",&w[i][j]) != 1) {
	      printf ("i = %d, j = %d\n",i,j);
	      utmc_msg ("Problem reading 2d matrix at i,j");
	      return 0;
	      }
	   }
   fclose (the_file);
   return 1;
   }

short write_3d_matrix (float ***w,int l_rows,int l_cols,int n_inp,
   char *file_name)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int i,j,k;
   the_file = fopen (file_name,"wt"); /* Stream file, write only, text */
   if (the_file == 0) return 0;
   fprintf (the_file,"%d  %d  %d\n",l_rows,l_cols,n_inp);
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++) {
	    for (k = 1;k <= n_inp;k++)
		fprintf (the_file,"%f  ",w[i][j][k]);
	    fprintf (the_file,"\n");
	    }
   fclose (the_file);
   return 1;
   }

void print_3d_matrix (float ***w,int rows,int cols,int inp)
   {
   int i,j,k;
   fprintf (stdprn,"Size: %d X  %d X  %d\n",rows,cols,inp);
   for (i = 1;i <= rows;i++)
       for (j = 1;j <= cols;j++) {
	    fprintf (stdprn,"%2d, %2d  ",i,j);
	    for (k = 1;k <= inp;k++)
		fprintf (stdprn,"%7.4f  ",w[i][j][k]);
	    fprintf (stdprn,"\n");
	    }
   return;
   }

void print_4d_matrix (float ****w,int rows,int cols,int inp,int inp2)
   {
   int i,j,k,l;
   fprintf (stdprn,"Size: %d X %d X %d X %d\n",rows,cols,inp,inp2);
   for (i = 1;i <= rows;i++)
       for (j = 1;j <= cols;j++)
	    for (k = 1;k <= inp;k++) {
		fprintf (stdprn,"%2d, %2d, %2d  ",i,j,k);
		for (l = 1;l <= inp2;l++)
		fprintf (stdprn,"%7.4f  ",w[i][j][k][l]);
	    fprintf (stdprn,"\n");
	    }
   return;
   }

short read_3d_matrix (float ***w,char *file_name,int rows,int cols,int inp)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int l_rows,l_cols,l_inp;
   int i,j,k;
   /* temp is used to read in one line of the file */
   char temp [MAX_LINE_LEN];
   the_file = fopen (file_name,"rt"); /* Stream file, read only, text */
   if (the_file == NULL) {
      utmc_msg ("Problem opening 3d matrix file.");
      return 0;
      }
   fgets (temp,MAX_LINE_LEN,the_file);
   if (sscanf (temp,"%d  %d  %d",&l_rows,&l_cols,&l_inp) != 3) {
      utmc_msg ("Problem reading dimensions of 3d matrix.");
      return 0;
      }
   if ((l_rows != rows)||(l_cols != cols)||(l_inp != inp)) {
      utmc_msg ("3d matrix file rows and cols do not equal setup values.");
      return 0;
      }
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++) {
	    for (k = 1;k <= l_inp;k++)
		if (fscanf (the_file,"%f",&w[i][j][k]) != 1) {
		    printf ("i = %d, j = %d, k = %d\n",i,j,k);
		    utmc_msg ("Problem reading 3d matrix entry at i,j,k");
		    return 0;
		    }
	    fgets (temp,MAX_LINE_LEN,the_file); /* read to end of line */
	    }
   fclose (the_file);
   return 1;
   }

short write_4d_matrix (float ****w,int l_rows,int l_cols,int m_inp,
   int l_inp,char *file_name)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int i,j,k,l;
   the_file = fopen (file_name,"wt"); /* Stream file, write only, text */
   if (the_file == 0) return 0;
   fprintf (the_file,"%d  %d  %d  %d\n",l_rows,l_cols,m_inp,l_inp);
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++)
	    for (k = 1;k <= m_inp;k++) {
		for (l = 1;l <= l_inp;l++)
		fprintf (the_file,"%f  ",w[i][j][k][l]);
		fprintf (the_file,"\n");
		}
   fclose (the_file);
   return 1;
   }

short read_4d_matrix (float ****w,char *file_name,int rows,int cols,
   int inp,int inp2)
   {
   /* Returns 0 if unsuccessful, 1 if successful  */
   FILE *the_file;
   int l_rows,l_cols,l_inp,l_inp2;
   int i,j,k,l;
   /* temp is used to read in one line of the file */
   char temp [MAX_LINE_LEN];
   the_file = fopen (file_name,"rt"); /* Stream file, read only, text */
   if (the_file == NULL) {
      utmc_msg ("Problem opening 4d matrix file.");
      return 0;
      }
   fgets (temp,MAX_LINE_LEN,the_file);
   if (sscanf (temp,"%d  %d  %d  %d",&l_rows,&l_cols,&l_inp,&l_inp2) != 4) {
      utmc_msg ("Problem reading dimensions of 4d matrix.");
      return 0;
      }
   if ((l_rows != rows)||(l_cols != cols)||(l_inp != inp)||(l_inp2 != inp2)) {
      utmc_msg ("4d matrix file rows and cols do not equal setup values.");
      return 0;
      }
   for (i = 1;i <= l_rows;i++)
       for (j = 1;j <= l_cols;j++)
	    for (k = 1;k <= l_inp;k++) {
		for (l = 1;l <= l_inp2;l++)
		if (fscanf (the_file,"%f",&w[i][j][k][l]) != 1) {
		    printf ("i = %d, j = %d, k = %d, l= %d\n",i,j,k,l);
		    utmc_msg("Problem reading 4d matrix entry at i,j,k,l");
		    return 0;
		    }
	    fgets (temp,MAX_LINE_LEN,the_file); /* read to end of line */
	    }
   fclose (the_file);
   return 1;
   }

#endif
